##
# aims
# lib/api.pl
# Copyright (c) 2012 Matt Ferris
# Author: Matt Ferris
#


##
# aimsapi_registerHandler
#
# $cmd The name of the clause
# $handler The sub to call
# $minarg Minimum number of arguments expected
# $maxarg Maximum number of arguments expected
#
# Allows modules to register a handler for a clause.
#
sub aimsapi_registerHandler
{
	my ($cmd, $handler, $minarg, $maxarg) = @_;
	$handlers->{$cmd} = {
		'handler' => 'aimsmod_'.$handler,
		'minarg' => $minarg,
		'maxarg' => $maxarg
		};
}

##
# aimsapi_registerAccessMethod
#
# $name Name of the access method
# $handler The sub to call
#
# Allows modules to register variable access
# methods.
#
sub aimsapi_registerAccessMethod
{
	my ($name, $handler) = @_;
	$accessMethods->{$name} = $handler
}

##
# aimsapi_pushBucket
#
# $b Name of the bucket (table, expr, chain, target)
# $data The data to add to the bucket
#
# Adds data to the currents rule's named bucket.
#
sub aimsapi_pushBucket
{
	my ($b, $data) = @_;
	push(@{$bucket->{$b}}, $data);
}

##
# aimsapi_setBucket
#
# $b Name of the bucket (table, expr, chain, target)
# $data The data to set the bucket to
#
# Sets the contents of the current rule's bucket
# section.
#
sub aimsapi_setBucket
{
	my ($b, $data) = @_;
	$bucket->{$b} = [$data];
}

##
# aimsapi_getBucket
#
# $b Name of the bucket
#
# Gets the contents of the current rule's bucket
#
sub aimsapi_getBucket
{
	my ($b) = @_;
	return $bucket->{$b};
}

##
# aimsapi_addRulesAtOffset
#
# $rulelist Array of rules to add
# $offset Position in the ruleset to add the rules to
#
# Add a list of rules to the ruleset at the specified
# offset.
#
sub aimsapi_addRulesAtOffset
{
	my ($rulelist, $offset) = @_;
	if ($offset == undef) { $offset = $#{$output} + 1; }

	my $tmprules = [];
	for (my $i; $i<=$#{$rules}; $i++)
	{
		if (($i + 1) == $offset)
		{
			foreach my $r (@$rulelist)
			{
				push(@$tmprules, $r);
			}
		}
		else
		{
			push(@$tmprules, $rules->[$i]);
		}
	}

	$rules = $tmprules;
}

##
# aimsapi_processFile
#
# $infile The filename to process
#
# Process the contents of a file.
#
sub aimsapi_processFile
{
	my ($infile) = @_;
	my $rules = [];
	my $tmpout = [];

	# debug level 1
	aims_debug(1, "api: start file $infile");

	push(@$file, $infile);

	open(IN, $infile) || aims_err($E_FATAL_NOFILE, $infile);
	while (<IN>)
	{
		next if /^$/;
		next if /^#/;
		my $rule = aims_tokenize($_);
		push (@$rules, $rule);
	}
	close(IN);

	aims_replvars($rules);
	aims_permute($rules);

	aims_compile($rules);

	pop(@$file);

	# debug level 1
	aims_debug(1, "api: end file $infile");
}

##
# aimsapi_isInterface
#
# $value The value to test
#
# Test if value is the name of a defined interface.
#
sub aimsapi_isInterface
{
	my ($value) = @_;
	my $if, $mod;

	if ($#{%$interfaces} <= 0)
	{
		$interfaces = aims_getifs();
	}

	if ($value =~ /^(eth[0-9\.\:]+|lo|tun[0-9\.\:]+)(\+(.+)|)/)
	{
		$if = $1;
	}

	if ($interfaces->{$if})
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

##
# aimsapi_getInterfaceValue
#
# $ifname Name of the interface
#
# Return the value for the given interface
#
sub aimsapi_getInterfaceValue
{
	my ($ifname) = @_;
	my $if, $mod, $value;

	if ($#{%$interfaces} <= 0)
	{
		$interfaces = aims_getifs();
	}

	if ($ifname =~ /^(eth[0-9\.\:]+|lo|tun[0-9\.\:]+)(\+(.+)|)/)
	{
		$if = $1;
		$mod = $3;

		if ($interfaces->{$if})
		{
			if ($mod)
			{
				if ($mod eq 'network')
				{
					$value = $interfaces->{$if}->{'ip'};
					$value .= '/'.$interfaces->{$if}->{'mask'};
				}
				else
				{
					aims_err($E_IFMOD_NOTDEF, $mod);
				}
			}
			else
			{
				$value = $interfaces->{$if}->{'ip'};
			}
		}

		aims_debug(2, "api: found value '$value' for '$ifname'");
	}

	return $value;
}

return 1;
