##
# aims
# list/core.pl
# Copyright (c) 2012 Matt Ferris
# Author: Matt Ferris
#

$handlers = {};
$variables = {};
$interfaces = {};
$bucket = {};
$accessMethods = {};
$file = [];
$output = [];

##
# aims_tokenize
#
# $rule The rule (in string form) to tokenize
# $delim The delimiter used to split on
#
# Divide a rule into tokens, splitting the rule
# up based on the specified delimiter (whitespace)
# by default, and return the result.
#
sub aims_tokenize
{
	my $rule = shift;
	my $delim = shift;

	chomp($rule);
	if (!$delim) { $delim = '\s'; }

	my $quotes = 0;
	my $tokens = [];
	my $token = '';

	# debug level 3
	aims_debug(3, "tokenize: start $rule");

	for (my $i=0; $i<=length($rule); $i++)
	{
		my $c = substr($rule, $i, 1);
		if ($c eq '"')
		{
			if ($quotes == 1) { $quotes = 0; }
			else { $quotes = 1; }
			$token .= $c;
		}
		elsif ($c =~ /$delim/ && $quotes == 0 && $enclosure == 0)
		{
			# debug level 3
			aims_debug(3, "tokenize: token: $token");

			push(@$tokens, $token);
			$token = '';
		}
		else
		{
			$token .= $c;
		}
	}
	if ($token ne '') 
	{
		# debug level 3
		aims_debug(3, "tokenize: last token: $token");

		push(@$tokens, $token);
	}

	# debug level 3
	aims_debug(3, "tokenize: end");

	return $tokens;
}

##
# aims_permute
#
# $rules The list of rules to permutate
#
# Expand value lists in the ruleset, creating a single
# rule for each value. Hand off rule processing to
# registered modules as well.
#
sub aims_permute
{
	my $rules = shift;
	my $inbuf = [];
	my $outbuf = [];
	push(@$inbuf, shift(@$rules));
	PRMLOOP: while (my $rule = shift(@$inbuf))
	{
		# debug level 2
		aims_debug(2, "permute: start @$rule");
		aims_debug(2, "permute: ".($#{$inbuf}+1)." rules in buffer");

		foreach my $c (keys(%$handlers))
		{
			if ($c =~ /^permute_/)
			{
				$newrule = aims_calhndlr($c, $rule, 0, 0);
				if ($newrule ne '')
				{
					push(@$inbuf, $rule);
					push(@$inbuf, $newrule);
					next PRMLOOP;
				}
			}
		}
		my $i = 0;
		foreach my $t (@$rule)
		{
			if ($t =~ /,/)
			{
				my $parts = aims_tokenize($t, ',');
				if ($#{$parts} <= 0) { last; }

				# debug level 2
				aims_debug(2, "permute: tokens: @$parts");

				foreach my $p (@$parts)
				{
					my $newrule = [];
					for (my $a=0; $a<$i; $a++) { push(@$newrule, $rule->[$a]); }
					push(@$newrule, $p);
					for (my $a=$i+1; $a<=$#{$rule}; $a++) { push(@$newrule, $rule->[$a]); }

					# debug level 2
					aims_debug(2, "permute: inbuf: @$newrule");

					push(@$inbuf, $newrule);
				}
				next PRMLOOP;
			}
			$i++;
		}
		push(@$outbuf, $rule);
		if ($#{$inbuf} == -1) { push(@$inbuf, shift(@$rules)); }

		# debug level 2
		aims_debug(2, "permute: end");
	}

	foreach my $o (@$outbuf) { push(@$rules, $o); }
}

##
# aims_replvar
#
# $rules The list of rules
# 
# Identify variable declarations, store the values
# and substitue them throughout the ruleset where
# required.
#
sub aims_replvars
{
	my $rules = shift;
	my $out = [];
	while ($#{$rules} > -1)
	{
		my $rule = shift(@$rules);
		if ($rule->[1] eq '=')
		{
			if ($rule->[2] =~ /^([a-zA-Z]+):(.+)$/)
			{
				my $am = $1;
				my $value = $2;

				# debug level 3
				aims_debug(3, "replvars: access method $am for $value");

				if ($accessMethods->{$am})
				{
					my $sub = 'aimsmod_'.$accessMethods->{$am};
					$variables->{$rule->[0]} = &$sub($value);
				}
			}
			else
			{
				$variables->{$rule->[0]} = $rule->[2];
			}
			next;
		}
		push(@$out, $rule);
	}

	foreach my $rule (@$out)
	{
		# debug level 3
		aims_debug(3, "replvars: start with @$rule");

		foreach my $t (@$rule)
		{
			# replace variables
			while ($t =~ /\$([a-zA-Z_][a-zA-Z0-9_]*)/)
			{
				my $name = $1;
				if ($variables->{$name} eq '') { aims_err($E_VAR_NOTDEF, $name, $rule); }
				my $value = $variables->{$name};

				# debug level 2
				aims_debug(2, "replvars: found variable $name, replacing with $value");

				$t =~ s/\$$name/$value/g;
			}
		}

		# debug level 3
		aims_debug(3, "replvars: end with @$rule");

		push(@$rules, $rule);
	}
}


##
# aims_compile
#
# $rules The list of rules
#
# Process the rules, passing off to registered
# modules where necessary, converting the rules
# to command-line calls to iptables. These calls
# are returned as a list.
#
sub aims_compile
{
	my $rules = shift;
	my $curcmd = '';
	my $lastcmd = '';
	my $tmpoutput = [];
	my $args = [];

	for (my $rulepos = 0; $rulepos<=$#{$rules}; $rulepos++)
	{
		my $tokens = $rules->[$rulepos];
		next if $#{$tokens} == 0;
		$bucket = aims_newbucket();
		$args = [];
		$curcmd = $lastcmd = '';

		# debug level 2
		aims_debug(2, "compile: start @$tokens");

		TKNLOOP: for (my $i = 0; $i<=$#{$tokens}; $i++)
		{
			my $t = $tokens->[$i];

			# debug level 2
			aims_debug("compile: token: $t");

			foreach my $c (keys(%$handlers))
			{
				my $minarg = $handlers->{$curcmd}->{'minarg'};
				my $maxarg = $handlers->{$curcmd}->{'maxarg'};
				next if $c eq '';

				# debug level 3
				aims_debug(3, "compile: token match: $t == $c, arg ".($#{$args}+1)." of $minarg to $maxarg");

				if ($curcmd && $t eq $c && ($#{$args}+1) >= $minarg && ($#{$args}+1) <= $maxarg)
				{
					if ($curcmd)
					{
						# debug level 2
						aims_debug(2, "compile: $curcmd with args: @$args");

						aims_calhndlr($curcmd, $args, $lastcmd, $rulepos);
						$args = [];
						$lastcmd = $curcmd;
					}
					$curcmd = $t;
					next TKNLOOP;
				}
				elsif (!$curcmd && $t eq $c)
				{
					# debug level 2
					aims_debug(2, "compile: found keyword $t");

					$curcmd = $t;
					next TKNLOOP;
				}
				elsif (($#{$args}+1) > $maxarg)
				{
					aims_err($E_PARSE_GENERAL, 'unexpected token: '.$tokens->[$i-1], $rules->[$rulepos]);
				}
			}
			if ($t =~ /^\"(.+)\"$/) { $t = $1; }

			# debug level 3
			aims_debug(3, "compile: token save: $t");

			push(@$args, $t);
		}

		# debug level 2
		aims_debug(2, "compile: last: $curcmd with args: @$args");

		if (($#{$args}+1) < $handlers->{$curcmd}->{'minarg'})
		{
			aims_err($E_PARSE_GENERAL, 'unexpected end of rule', $rules->[$rulepos]);
		}

		aims_calhndlr($curcmd, $args, $lastcmd, $rulepos);

		# debug level 2
		aims_debug(2, "compile: end");

		if ($bucket->{'table'} ne '')
		{
			my $out =
				join(' ', @{$bucket->{'table'}}).' '.
				join(' ', @{$bucket->{'chain'}}).' '.
				join(' ', @{$bucket->{'expr'}}).' '.
				join(' ', @{$bucket->{'target'}});
			undef($bucket);

			unless ($out =~ /^\s+$/)
			{
				$out =~ s/\s{2,}/ /g;

				# debug level 3
				aims_debug(3, "compile: push $out");

				push(@$output, $out);
			}
		}
	}
}

##
# aims_loadmods
#
# $dir The directory to search for modules
#
# Search the specified directory for modules,
# and load any that are found.
#
sub aims_loadmods
{
	my ($dir) = @_;
	opendir(DIR, $dir) || die("failed to read mods directory '$dir': $!\n");
	while ($file = readdir(DIR))
	{
		next unless $file =~ /^.+\.pl$/;
		require $dir.'/'.$file;
		$file =~ s/\.pl//;
		$initsub = 'aimsmod_'.$file.'_init';
		&$initsub();
	}
	closedir(DIR);
}

##
# aims_calhndlr
#
# $cmd The name of the clause
# $args The list of arguments following the clause
# $lastcmd The last clause processed
# $rulepos The position of the current rule in the ruleset
#
# Call the handler(s) for the current clause.
#
sub aims_calhndlr
{
	my $cmd = shift;
	my $args = shift;
	my $lastcmd = shift;
	my $rulepos = shift;
	if (!defined($handlers->{$cmd}->{'handler'})) { aims_err($E_HANDLER_NOEXIST, $cmd, $rules->[$rulepos]); }
	if (!$handlers->{$cmd}->{'handler'}) { aims_err($E_PARSE_NOHNDLR, $cmd, $rules->[$rulepos]); }
	my $sub = $handlers->{$cmd}->{'handler'};
	&$sub($args, $lastcmd, $rulepos);
}

##
# aims_newbucket
#
# Return a new bucket
#
sub aims_newbucket
{
	my $bucket = {
		'table' => [],
		'chain' => [],
		'expr' => [],
		'target' => []
	};
	return $bucket;
}

##
# aims_printmsg
#
# $msg The message to print
#
# Prints a message to stdout
#
sub aims_printmsg
{
	my $msg = shift;
	my $curfile = $file->[$#{$file}];
	if ($curfile ne '') { print "$curfile: "; }
	print "$msg\n";
}

##
# aims_debug
#
# $level The debug level of the message
# $msg The debugging message
#
# Sends a debugging message of a specified level
#
sub aims_debug
{
	my $level = shift;
	my $msg = shift;
	if ($level != 0 && $verbose >= $level)
	{
		aims_printmsg($msg);
	}
}

##
# aims_getifs
#
# Return a list of interfaces and their corresponding
# IP addresses.
#
sub aims_getifs
{
	my $ifMarkers = {};
	my $currentif;
	my @lines = `ifconfig -a`;
	foreach (@lines)
	{
		chomp();
		if (/^[a-zA-Z]/)
		{
			my ($ifname, @other) = split(/\s/);
			$currentif = $ifname;
			$ifMarkers->{$currentif} = { 'ip' => '', 'mask' => '255.255.255.255'};
		}
		if (/inet addr:([a-zA-Z0-9_\:\.]+)(\s+Bcast:[0-9\.]+|)\s+Mask:([0-9\.]+)/)
		{
			$ip = $1;
			$mask = $3;
			$ifMarkers->{$currentif}->{'ip'} = $ip;
			$ifMarkers->{$currentif}->{'mask'} = $mask;
		}
	}
	return $ifMarkers;
}

return 1;
